@@ -5,6 +5,7 @@ from decimal import Decimal
 
 from simpleeval import simple_eval
 
+from trytond.i18n import gettext
 from trytond.pool import PoolMeta, Pool
 from trytond.model import (
     ModelSQL, ModelView, MatchMixin, Workflow, DeactivableMixin, fields)
@@ -12,10 +13,8 @@ from trytond.pyson import Eval, If, Bool
 from trytond.transaction import Transaction
 from trytond.tools import decistmt
 
-from trytond.modules.product import price_digits
-
-__all__ = ['Sale', 'SaleLine',
-    'SalePromotion', 'SalePromotion_Product', 'SalePromotion_ProductCategory']
+from trytond.modules.product import price_digits, round_price
+from .exceptions import FormulaError
 
 
 class Sale(metaclass=PoolMeta):
@@ -68,7 +67,7 @@ class Sale(metaclass=PoolMeta):
             promotion.apply(self)
 
 
-class SaleLine(metaclass=PoolMeta):
+class Line(metaclass=PoolMeta):
     __name__ = 'sale.line'
 
     draft_unit_price = fields.Numeric('Draft Unit Price',
@@ -80,21 +79,17 @@ class SaleLine(metaclass=PoolMeta):
             ])
 
 
-class SalePromotion(
+class Promotion(
         DeactivableMixin, ModelSQL, ModelView, MatchMixin):
     'Sale Promotion'
     __name__ = 'sale.promotion'
 
     name = fields.Char('Name', translate=True, required=True)
-    company = fields.Many2One('company.company', 'Company', required=True,
+    company = fields.Many2One(
+        'company.company', "Company", required=True, select=True,
         states={
             'readonly': Eval('id', 0) > 0,
-            },
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ],
-        select=True)
+            })
     start_date = fields.Date('Start Date',
         domain=['OR',
             ('start_date', '<=', If(~Eval('end_date', None),
@@ -117,6 +112,25 @@ class SalePromotion(
             ('company', '=', Eval('company', -1)),
             ],
         depends=['company'])
+
+    amount = fields.Numeric(
+        "Amount", digits=(16, Eval('currency_digits', 2)),
+        depends=['currency_digits'])
+    currency = fields.Many2One(
+        'currency.currency', "Currency",
+        states={
+            'required': Bool(Eval('amount', 0)),
+            },
+        depends=['amount'])
+    currency_digits = fields.Function(
+        fields.Integer("Currency Digits"), 'on_change_with_currency_digits')
+    untaxed_amount = fields.Boolean(
+        "Untaxed Amount",
+        states={
+            'invisible': ~Eval('amount'),
+            },
+        depends=['amount'])
+
     quantity = fields.Float('Quantity', digits=(16, Eval('unit_digits', 2)),
         depends=['unit_digits'])
     unit = fields.Many2One('product.uom', 'Unit',
@@ -126,37 +140,45 @@ class SalePromotion(
         depends=['quantity'])
     unit_digits = fields.Function(fields.Integer('Unit Digits'),
         'on_change_with_unit_digits')
-    products = fields.Many2Many('sale.promotion-product.product',
-        'promotion', 'product', 'Products')
+    products = fields.Many2Many(
+        'sale.promotion-product.product', 'promotion', 'product', "Products",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     categories = fields.Many2Many(
         'sale.promotion-product.category', 'promotion', 'category',
-        "Categories")
+        "Categories",
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     formula = fields.Char('Formula', required=True,
         help=('Python expression that will be evaluated with:\n'
             '- unit_price: the original unit_price'))
 
-    @classmethod
-    def __setup__(cls):
-        super(SalePromotion, cls).__setup__()
-        cls._error_messages.update({
-                'invalid_formula': ('Invalid formula "%(formula)s" '
-                    'in promotion "%(promotion)s" '
-                    'with exception "%(exception)s".'),
-                })
-
     @staticmethod
     def default_company():
         return Transaction().context.get('company')
 
+    @classmethod
+    def default_untaxed_amount(cls):
+        return False
+
     @fields.depends('unit')
     def on_change_with_unit_digits(self, name=None):
         if self.unit:
             return self.unit.digits
         return 2
 
+    @fields.depends('currency')
+    def on_change_with_currency_digits(self, name=None):
+        if self.currency:
+            return self.currency.digits
+
     @classmethod
     def validate(cls, promotions):
-        super(SalePromotion, cls).validate(promotions)
+        super().validate(promotions)
         for promotion in promotions:
             promotion.check_formula()
 
@@ -166,11 +188,11 @@ class SalePromotion(
             if not isinstance(self.get_unit_price(**context), Decimal):
                 raise ValueError('Not a Decimal')
         except Exception as exception:
-            self.raise_user_error('invalid_formula', {
-                    'formula': self.formula,
-                    'promotion': self.rec_name,
-                    'exception': exception,
-                    })
+            raise FormulaError(
+                gettext('sale_promotion.msg_invalid_formula',
+                    formula=self.formula,
+                    promotion=self.rec_name,
+                    exception=exception)) from exception
 
     @classmethod
     def _promotions_domain(cls, sale):
@@ -208,24 +230,43 @@ class SalePromotion(
 
     def get_pattern(self, sale):
         pool = Pool()
+        Currency = pool.get('currency.currency')
         Uom = pool.get('product.uom')
+        Sale = pool.get('sale.sale')
         pattern = {}
-        if not self.unit:
-            return pattern
-        quantity = 0
-        for line in sale.lines:
-            if self.is_valid_sale_line(line):
-                quantity += Uom.compute_qty(line.unit, line.quantity,
-                    self.unit)
-        pattern['quantity'] = quantity
+        if self.currency:
+            amount = self.get_sale_amount(Sale(sale.id))
+            pattern['amount'] = Currency.compute(
+                sale.currency, amount, self.currency)
+        if self.unit:
+            quantity = 0
+            for line in sale.lines:
+                if self.is_valid_sale_line(line):
+                    quantity += Uom.compute_qty(line.unit, line.quantity,
+                        self.unit)
+            pattern['quantity'] = quantity
         return pattern
 
     def match(self, pattern):
+        def sign(amount):
+            return Decimal(1).copy_sign(amount)
         if 'quantity' in pattern:
             pattern = pattern.copy()
-            if self.quantity > pattern.pop('quantity'):
+            if (self.quantity or 0) > pattern.pop('quantity'):
                 return False
-        return super(SalePromotion, self).match(pattern)
+        if 'amount' in pattern:
+            pattern = pattern.copy()
+            amount = pattern.pop('amount')
+            if (sign(self.amount or 0) * sign(amount) >= 0
+                    and abs(self.amount or 0) > abs(amount)):
+                return False
+        return super().match(pattern)
+
+    def get_sale_amount(self, sale):
+        if self.untaxed_amount:
+            return sale.untaxed_amount
+        else:
+            return sale.total_amount
 
     def is_valid_sale_line(self, line):
 
@@ -249,21 +290,19 @@ class SalePromotion(
             return True
 
     def apply(self, sale):
-        new_prices = {}
+        applied = False
         for line in sale.lines:
             if line.type != 'line':
                 continue
             if not self.is_valid_sale_line(line):
                 continue
             context = self.get_context_formula(line)
-            new_prices[line] = self.get_unit_price(**context)
-
-        # Apply promotion only if all unit prices decrease
-        if all(l.unit_price >= p for l, p in new_prices.items()):
-            for line, unit_price in new_prices.items():
-                line.unit_price = unit_price.quantize(
-                    Decimal(1) / 10 ** line.__class__.unit_price.digits[1])
+            new_price = self.get_unit_price(**context)
+            if line.unit_price >= new_price:
+                line.unit_price = round_price(new_price)
                 line.promotion = self
+                applied = True
+        if applied:
             sale.lines = sale.lines  # Trigger the change
 
     def get_context_formula(self, sale_line):
@@ -288,7 +327,7 @@ class SalePromotion(
         return max(simple_eval(decistmt(self.formula), **context), Decimal(0))
 
 
-class SalePromotion_Product(ModelSQL):
+class Promotion_Product(ModelSQL):
     'Sale Promotion - Product'
     __name__ = 'sale.promotion-product.product'
 
@@ -298,7 +337,7 @@ class SalePromotion_Product(ModelSQL):
         required=True, ondelete='CASCADE')
 
 
-class SalePromotion_ProductCategory(ModelSQL):
+class Promotion_ProductCategory(ModelSQL):
     "Sale Promotion - Product Category"
     __name__ = 'sale.promotion-product.category'
 
